/*  OpenTally: Open-source election vote counting
 *  Copyright © 2021 Lee Yingtong Li (RunasSudo)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

mod utils;

use opentally::constraints::Constraints;
use opentally::election::{CandidateState, CountState, Election};
use opentally::numbers::Rational;
use opentally::stv;

use std::fs::File;
use std::io::{self, BufRead}; 

#[test]
fn prsa1_constr1_rational() {
	// FIXME: This is unvalidated!
	
	// Read BLT
	let file = File::open("tests/data/prsa1.blt").expect("IO Error");
	let file_reader = io::BufReader::new(file);
	let lines = file_reader.lines();
	let mut election: Election<Rational> = Election::from_blt(lines.map(|r| r.expect("IO Error").to_string()).into_iter());
	
	// Read CON
	let file = File::open("tests/data/prsa1_constr1.con").expect("IO Error");
	let file_reader = io::BufReader::new(file);
	let lines = file_reader.lines();
	election.constraints = Some(Constraints::from_con(lines.map(|r| r.expect("IO Error").to_string()).into_iter()));
	
	let stv_opts = stv::STVOptions {
		round_tvs: Some(3),
		round_weights: Some(3),
		round_votes: Some(3),
		round_quota: Some(3),
		sum_surplus_transfers: stv::SumSurplusTransfersMode::SingleStep,
		meek_surplus_tolerance: String::new(),
		normalise_ballots: false,
		quota: stv::QuotaType::Droop,
		quota_criterion: stv::QuotaCriterion::GreaterOrEqual,
		quota_mode: stv::QuotaMode::Static,
		ties: vec![],
		surplus: stv::SurplusMethod::EG,
		surplus_order: stv::SurplusOrder::ByOrder,
		transferable_only: true,
		exclusion: stv::ExclusionMethod::ParcelsByOrder,
		meek_nz_exclusion: false,
		early_bulk_elect: false,
		bulk_exclude: false,
		defer_surpluses: false,
		meek_immediate_elect: false,
		constraints_path: Some("tests/data/prsa1_constr1.con".to_string()),
		constraint_mode: stv::ConstraintMode::GuardDoom,
		pp_decimals: 2,
	};
	
	// Initialise count state
	let mut state = CountState::new(&election);
	
	// Count election
	stv::count_init(&mut state, &stv_opts).unwrap();
	while stv::count_one_stage::<Rational>(&mut state, &stv_opts).unwrap() == false {}
	
	// Validate winners
	let mut winners = Vec::new();
	for (candidate, count_card) in state.candidates.iter() {
		if count_card.state == CandidateState::Elected {
			winners.push((candidate, count_card));
		}
	}
	winners.sort_unstable_by(|a, b| a.1.order_elected.partial_cmp(&b.1.order_elected).unwrap());
	
	assert_eq!(winners[0].0.name, "Grey");
	assert_eq!(winners[1].0.name, "Ames");
	assert_eq!(winners[2].0.name, "White");
	assert_eq!(winners[3].0.name, "Evans");
}

#[test]
fn prsa1_constr2_rational() {
	// FIXME: This is unvalidated!
	
	// Read BLT
	let file = File::open("tests/data/prsa1.blt").expect("IO Error");
	let file_reader = io::BufReader::new(file);
	let lines = file_reader.lines();
	let mut election: Election<Rational> = Election::from_blt(lines.map(|r| r.expect("IO Error").to_string()).into_iter());
	
	// Read CON
	let file = File::open("tests/data/prsa1_constr2.con").expect("IO Error");
	let file_reader = io::BufReader::new(file);
	let lines = file_reader.lines();
	election.constraints = Some(Constraints::from_con(lines.map(|r| r.expect("IO Error").to_string()).into_iter()));
	
	let stv_opts = stv::STVOptions {
		round_tvs: Some(3),
		round_weights: Some(3),
		round_votes: Some(3),
		round_quota: Some(3),
		sum_surplus_transfers: stv::SumSurplusTransfersMode::SingleStep,
		meek_surplus_tolerance: String::new(),
		normalise_ballots: false,
		quota: stv::QuotaType::Droop,
		quota_criterion: stv::QuotaCriterion::GreaterOrEqual,
		quota_mode: stv::QuotaMode::Static,
		ties: vec![],
		surplus: stv::SurplusMethod::EG,
		surplus_order: stv::SurplusOrder::ByOrder,
		transferable_only: true,
		exclusion: stv::ExclusionMethod::ParcelsByOrder,
		meek_nz_exclusion: false,
		early_bulk_elect: false,
		bulk_exclude: false,
		defer_surpluses: false,
		meek_immediate_elect: false,
		constraints_path: Some("tests/data/prsa1_constr2.con".to_string()),
		constraint_mode: stv::ConstraintMode::GuardDoom,
		pp_decimals: 2,
	};
	
	// Initialise count state
	let mut state = CountState::new(&election);
	
	// Count election
	stv::count_init(&mut state, &stv_opts).unwrap();
	while stv::count_one_stage::<Rational>(&mut state, &stv_opts).unwrap() == false {}
	
	// Validate winners
	let mut winners = Vec::new();
	for (candidate, count_card) in state.candidates.iter() {
		if count_card.state == CandidateState::Elected {
			winners.push((candidate, count_card));
		}
	}
	winners.sort_unstable_by(|a, b| a.1.order_elected.partial_cmp(&b.1.order_elected).unwrap());
	
	assert_eq!(winners[0].0.name, "Grey");
	assert_eq!(winners[1].0.name, "Ames");
	assert_eq!(winners[2].0.name, "White");
	assert_eq!(winners[3].0.name, "Evans");
}

#[test]
fn prsa1_constr3_rational() {
	// FIXME: This is unvalidated!
	
	// Read BLT
	let file = File::open("tests/data/prsa1.blt").expect("IO Error");
	let file_reader = io::BufReader::new(file);
	let lines = file_reader.lines();
	let mut election: Election<Rational> = Election::from_blt(lines.map(|r| r.expect("IO Error").to_string()).into_iter());
	
	// Read CON
	let file = File::open("tests/data/prsa1_constr3.con").expect("IO Error");
	let file_reader = io::BufReader::new(file);
	let lines = file_reader.lines();
	election.constraints = Some(Constraints::from_con(lines.map(|r| r.expect("IO Error").to_string()).into_iter()));
	
	let stv_opts = stv::STVOptions {
		round_tvs: Some(3),
		round_weights: Some(3),
		round_votes: Some(3),
		round_quota: Some(3),
		sum_surplus_transfers: stv::SumSurplusTransfersMode::SingleStep,
		meek_surplus_tolerance: String::new(),
		normalise_ballots: false,
		quota: stv::QuotaType::Droop,
		quota_criterion: stv::QuotaCriterion::GreaterOrEqual,
		quota_mode: stv::QuotaMode::Static,
		ties: vec![],
		surplus: stv::SurplusMethod::EG,
		surplus_order: stv::SurplusOrder::ByOrder,
		transferable_only: true,
		exclusion: stv::ExclusionMethod::ParcelsByOrder,
		meek_nz_exclusion: false,
		early_bulk_elect: false,
		bulk_exclude: false,
		defer_surpluses: false,
		meek_immediate_elect: false,
		constraints_path: Some("tests/data/prsa1_constr2.con".to_string()),
		constraint_mode: stv::ConstraintMode::GuardDoom,
		pp_decimals: 2,
	};
	
	// Initialise count state
	let mut state = CountState::new(&election);
	
	// Count election
	stv::count_init(&mut state, &stv_opts).unwrap();
	while stv::count_one_stage::<Rational>(&mut state, &stv_opts).unwrap() == false {}
	
	// Validate winners
	let mut winners = Vec::new();
	for (candidate, count_card) in state.candidates.iter() {
		if count_card.state == CandidateState::Elected {
			winners.push((candidate, count_card));
		}
	}
	winners.sort_unstable_by(|a, b| a.1.order_elected.partial_cmp(&b.1.order_elected).unwrap());
	
	assert_eq!(winners[0].0.name, "Grey");
	assert_eq!(winners[1].0.name, "White");
	assert_eq!(winners[2].0.name, "Thomson");
	assert_eq!(winners[3].0.name, "Reid");
}
